/* pjpeg.c */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#ifdef __MINGW32__
#ifdef INT32
#undef INT32
#endif
#include <windows.h>
#endif

#include "print.h"
#include "jpeglib.h"

struct my_error_mgr
{
 struct jpeg_error_mgr pub; /* "public" fields */
 jmp_buf setjmp_buffer;	    /* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

static struct jpeg_compress_struct cinfo;
static struct my_error_mgr jerr;

static FILE *fout=NULL;
static char *fname=NULL;
static int printModeType;

/* Here's the routine that will replace the standard error_exit method: */
METHODDEF(void) my_error_exit (j_common_ptr cinfo)
{
 /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
 my_error_ptr myerr = (my_error_ptr) cinfo->err;
 /* Always display the message. */
 /* We could postpone this until after returning, if we chose. */
 (*cinfo->err->output_message) (cinfo);
 /* Return control to the setjmp point */
 longjmp(myerr->setjmp_buffer, 1);
}

static int initPrint(int printType, int complexImage, int xpixels,
	int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
 int filedes = -1;

 printModeType = printType;

 if (fname)
  fname=NULL;
 if (fout)
  fclose(fout);
#ifdef __MINGW32__
 char tempPath[MAX_PATH + 1];

 if (GetTempPathA(MAX_PATH + 1, tempPath) == 0)
  {
   // Failed to get temp path.
   perror("plib: failed to get temp path");
   return (PRT_IO_ERROR);
  }

 // This is allocated in advance because of the code below.
 char tempFilename[MAX_PATH + 1];

 fname = tempFilename;
 if (GetTempFileNameA(tempPath, "prt", 1, fname) == 0)
  {
   perror("plib: failed to create tmpfile");
   return(PRT_IO_ERROR);
  }

 fout = fopen(fname, "wb");
 fname = strdup(fname);
#else
 fname = strdup("/tmp/prtXXXXXX");
 if ((filedes=mkstemp(fname)) < 0 )
  return(PRT_IO_ERROR);
 else
  {
   /* instead of fout = fdopen(filedes, "wb"); */
   /* close and reopen the file as fclose does not close files created by mkstemp */
   close(filedes);
   fout = fopen(fname, "wb");
  }
#endif


 /* just a paranoid check; should not really happen */
 if (printModeType!=IMG_TRUE_COL)
  return(PRT_INTERNAL_ERROR);

 /* set up the error handler */
 cinfo.err = jpeg_std_error(&jerr.pub);
 jerr.pub.error_exit = my_error_exit;
 if (setjmp(jerr.setjmp_buffer))
  return(PRT_INTERNAL_ERROR);

 /* initialize the JPEG compression object. */
 jpeg_create_compress(&cinfo);
 /* set the output file */
 jpeg_stdio_dest(&cinfo, fout);

 cinfo.image_width = xpixels;   /* image width and height, in pixels */
 cinfo.image_height = ypixels;
 cinfo.input_components = 3;    /* # of color components per pixel */
 cinfo.in_color_space = JCS_RGB;/* colorspace of input image */
 /* Now use the library's routine to set default compression parameters.
  * (You must set at least cinfo.in_color_space before calling this,
  * since the defaults depend on the source color space.)
  */
 jpeg_set_defaults(&cinfo);
 /* Now you can set any non-default parameters you wish to.
  * Here we just illustrate the use of quality (quantization table) scaling:
  */
 jpeg_set_quality(&cinfo, 80, TRUE); /* limit to baseline-JPEG values */

 /* disable the default 2x2 chroma subsampling */
 cinfo.comp_info[0].h_samp_factor = 1;
 cinfo.comp_info[0].v_samp_factor = 1;
 cinfo.comp_info[1].h_samp_factor = 1;
 cinfo.comp_info[1].v_samp_factor = 1;
 cinfo.comp_info[2].h_samp_factor = 1;
 cinfo.comp_info[2].v_samp_factor = 1;

 cinfo.write_JFIF_header = TRUE;
 cinfo.density_unit = 1; /* dots per inch */
 cinfo.X_density = xdpi/1000;
 cinfo.Y_density = ydpi/1000;
 jpeg_start_compress(&cinfo, TRUE);

 return(PRT_OK);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
	int lines, char bandType, unsigned char *cmap)
{
 int x, y;
 JSAMPROW row_pointer[lines];	/* pointer to JSAMPLE row[s] */
 unsigned char data2[3*MAX_PRINT], *data2ptr, *cmapptr;

 if (setjmp(jerr.setjmp_buffer))
  return(PRT_INTERNAL_ERROR);

 if (data==NULL)
  return(PRT_INTERNAL_ERROR);
 if (printModeType<bandType)
  return(PRT_WRONG_IMG);

 if (printModeType!=IMG_TRUE_COL)
  return(PRT_INTERNAL_ERROR);

 if (bandType>=IMG_TRUE_COL_DITH)
  {
   for (y=0; y<lines; y++)
    row_pointer[y] = (JSAMPROW) &data[y*bpl];
   (void) jpeg_write_scanlines(&cinfo, row_pointer, lines);
  }
 else /* we need to upgrade to 24 bits per pixel */
  for (y=0; y<lines; y++)
   {
    row_pointer[0] = data2ptr = data2;
    if (bandType==IMG_BW)
     {
      for (x=0; x<xpixels; x++)
       if (data[y*bpl+(x>>3)]&(128>>(x&7)))
        { *data2ptr++=0; *data2ptr++=0; *data2ptr++=0; }
       else
        { *data2ptr++=255; *data2ptr++=255; *data2ptr++=255; }
     }
    else /* must be IMG_CMAP */
     {
      for (x=0; x<xpixels; x++)
       { cmapptr = &cmap[((unsigned char)data[y*bpl+x])*3];
         *data2ptr++ = *cmapptr++; *data2ptr++ = *cmapptr++; *data2ptr++ = *cmapptr; }
     }
   (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }
 return(PRT_OK);
}

static int endPrint(int printStatus)
{
 if (setjmp(jerr.setjmp_buffer)); /* do nothing on purpose */
 else
  jpeg_finish_compress(&cinfo);
 if (printStatus==PRT_OK)
  {
   if (fout)
    {
     fflush(fout);
     fclose(fout);
     fout=NULL;
    }
   printStatus = enqueuePrintFile(fname);
  }
 else
  {
   if (fout)
    {
     fclose(fout);
     fout=NULL;
    }
  }
 /* just cleanup; called even if printing fails */
 jpeg_destroy_compress(&cinfo);
 if (fname)
  {
   removeFileFreeFilename(fname, printStatus);
   fname=NULL;
  }
 return(printStatus);
}

/* fill in the preallocated printer modes for JPEG */
static void getPrtInfo(printerMode *printerMod, int *numModes,
	printerDensity *prtDensityX, int *numDensitX,
	printerDensity *prtDensityY, int *numDensitY,
	int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
	int *dpiVariable, char **prtName, char **profileName)
{
 int i;

 printerMod[0].printType = IMG_TRUE_COL;
 *numModes=1;
 *numDensitX = *numDensitY = 7;
 for (i=0; i<6; i++)
  {
   prtDensityX[i].pixels = prtDensityY[i].pixels = MAX_PRINT;
   prtDensityX[i].dpi = prtDensityY[i].dpi = 100000*(i+1);
  }
 prtDensityX[6].pixels = prtDensityY[6].pixels = MAX_PRINT;
 prtDensityX[6].dpi = prtDensityY[6].dpi = 720000;

 *nonprintableTop=0;
 *nonprintableBottom=0;
 *nonprintableLeft=0;
 *dpiVariable=TRUE;
 *prtName="JPEG";
 *profileName=NULL;
}

/* register JPEG "printer" */
void registerJPEG()
{
 registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint);
}
